import 'package:flutter_webrtc/media_stream.dart';
import 'package:flutter_webrtc/webrtc.dart';
import 'package:xmpp_stone/xmpp_stone.dart';

import 'package:connectycube_sdk/connectycube_chat.dart';

import 'p2p_client.dart';
import '../call_session.dart';
import '../models/call_base_session.dart';
import '../models/cube_rtc_session_description.dart';
import '../peer_conntection.dart';
import '../utils/rtc_media_config.dart';
import '../utils/signaling_messages_utils.dart';
import '../utils/signaling_specifications.dart';

class P2PSession extends BaseSession
    implements CallSession<P2PSession>, PeerConnectionStateCallback {
  static const String TAG = "P2PSession";

  @override
  UserActionCallback<P2PSession> onCallAcceptedByUser;
  @override
  UserActionCallback<P2PSession> onCallRejectedByUser;
  @override
  LocalStreamCallback onLocalStreamReceived;
  @override
  UserActionCallback<P2PSession> onReceiveHungUpFromUser;
  @override
  RemoteStreamCallback<P2PSession> onRemoteStreamReceived;
  @override
  RemoteStreamCallback<P2PSession> onRemoteStreamRemoved;
  @override
  UserConnectionStateCallback<P2PSession> onUserNoAnswer;
  @override
  SessionClosedCallback<P2PSession> onSessionClosed;

  P2PClient _client;
  CubeRTCSessionDescription _cubeSdp;
  MediaStream _localStream;
  RTCSessionState _state;

  int get callerId => _cubeSdp.callerId;

  String get sessionId => _cubeSdp.sessionId;

  int get callType => _cubeSdp.conferenceType;

  Set<int> get opponentsIds => _cubeSdp.opponents;

  RTCSessionState get state => _state;

  Map<int, PeerConnection> _channels = {};

  P2PSession(this._client, this._cubeSdp) {
    _setState(RTCSessionState.RTC_SESSION_NEW);
    _createChannelsForOpponents(_cubeSdp.opponents);
  }

  void _createChannelsForOpponents(Set<int> opponentIds) {
    CubeUser currentUser = CubeChatConnection.instance.currentUser;

    opponentIds.forEach((opponentId) {
      if (opponentId != currentUser.id) {
        _channels[opponentId] = PeerConnection(opponentId, this);
      } else {
        _channels[callerId] = PeerConnection(callerId, this);
      }
    });
  }

  @override
  Future<void> acceptCall([Map<String, String> userInfo]) async {
    _setState(RTCSessionState.RTC_SESSION_CONNECTING);

    await _initLocalMediaStream();

    _channels.forEach((opponentId, peerConnection) {
      CubeUser currentUser = CubeChatConnection.instance.currentUser;

      if (opponentId == callerId) {
        peerConnection.startAnswer(userInfo);
      } else if (opponentId < currentUser.id) {
        peerConnection.startOffer(userInfo);
      } else if (peerConnection.hasRemoteSdp()) {
        peerConnection.startAnswer(userInfo);
      }
    });
  }

  @override
  void hungUp([Map<String, String> userInfo]) {
    _channels.forEach((opponentId, peerConnection) {
      _sendHungUpMessage(opponentId, userInfo);
      peerConnection.close();
    });

    _closeCurrentSession();
  }

  @override
  void reject([Map<String, String> userInfo]) {
    _channels.forEach((opponentId, peerConnection) {
      _sendRejectMessage(opponentId, userInfo);
      peerConnection.close();
    });

    _closeCurrentSession();
  }

  @override
  Future<void> startCall([Map<String, String> userInfo]) async {
    log("start call $userInfo");
    _setState(RTCSessionState.RTC_SESSION_CONNECTING);

    await _initLocalMediaStream();

    _channels.forEach((opponentId, peerChannel) {
      peerChannel.startOffer(userInfo);
    });
  }

  @override
  Future<bool> switchCamera() {
    if (CallType.VIDEO_CALL != callType)
      return Future.error(
          IllegalStateException("Can't perform operation for AUDIO call"));

    try {
      return _localStream?.getVideoTracks()[0]?.switchCamera();
    } catch (error) {
      return Future.error(error);
    }
  }

  @override
  void setVideoEnabled(bool enabled) {
    if (CallType.VIDEO_CALL != callType) return;

    _localStream?.getVideoTracks()[0].enabled = enabled;
  }

  @override
  void setMicrophoneMute(bool mute) {
    _localStream?.getAudioTracks()[0].setMicrophoneMute(mute);
  }

  @override
  void enableSpeakerphone(bool enable) {
    _localStream?.getAudioTracks()[0].enableSpeakerphone(enable);
  }

  Future<MediaStream> _createStream(bool isScreenSharing) async {
    final Map<String, dynamic> mediaConstraints = {
      'audio': true,
    };

    if (CallType.VIDEO_CALL == callType) {
      RTCMediaConfig mediaConfig = RTCMediaConfig.instance;
      mediaConstraints['video'] = {
        'mandatory': {
          'minWidth': mediaConfig.minWidth,
          'minHeight': mediaConfig.minHeight,
          'minFrameRate': mediaConfig.minFrameRate,
        },
        'facingMode': 'user',
        'optional': [],
      };
    }

    MediaStream stream = isScreenSharing
        ? await navigator.getDisplayMedia(mediaConstraints)
        : await navigator.getUserMedia(mediaConstraints);
    if (this.onLocalStreamReceived != null) {
      this.onLocalStreamReceived(stream);
    }
    return stream;
  }

  Future<void> _initLocalMediaStream() async {
    _localStream = await _createStream(false);
  }

  @override
  MediaStream getLocalMediaStream() => _localStream;

  @override
  void onRemoteStreamReceive(int userId, MediaStream remoteMediaStream) {
    if (onRemoteStreamReceived != null) {
      onRemoteStreamReceived(this, userId, remoteMediaStream);
    }
  }

  @override
  void onRemoteStreamRemove(int userId, MediaStream remoteMediaStream) {
    if (onRemoteStreamRemoved != null) {
      onRemoteStreamRemoved(this, userId, remoteMediaStream);
    }
  }

  @override
  void onSendAnswer(
    int userId,
    RTCSessionDescription sdp,
    Map<String, String> userInfo,
  ) {
    MessageStanza answerMessage =
        createAnswerMessage(_cubeSdp, userId, userInfo, sdp);
    _client.sendSignalingMessage(answerMessage);
  }

  @override
  void onSendOffer(
    int userId,
    RTCSessionDescription sdp,
    Map<String, String> userInfo,
  ) {
    MessageStanza offerMessage = createOfferMessage(
      _cubeSdp,
      userId,
      userInfo,
      sdp,
    );
    _client.sendSignalingMessage(offerMessage);
  }

  @override
  void sendIceCandidate(int userId, RTCIceCandidate iceCandidate) {
    log("sendIceCandidate, _sdp = $_cubeSdp");
    MessageStanza iceCandidatesMessage = createIceCandidatesMessage(
      _cubeSdp,
      userId,
      [iceCandidate],
    );
    _client.sendSignalingMessage(iceCandidatesMessage);
  }

  @override
  void onPeerConnectionStateChanged(int userId, PeerConnectionState state) {
    switch (state) {
      case PeerConnectionState.RTC_CONNECTION_NEW:
        break;

      case PeerConnectionState.RTC_CONNECTION_PENDING:
        break;

      case PeerConnectionState.RTC_CONNECTION_CHECKING:
        break;

      case PeerConnectionState.RTC_CONNECTION_CONNECTING:
        break;

      case PeerConnectionState.RTC_CONNECTION_CONNECTED:
        _setState(RTCSessionState.RTC_SESSION_CONNECTED);
        break;

      case PeerConnectionState.RTC_CONNECTION_DISCONNECTED:
        break;

      case PeerConnectionState.RTC_CONNECTION_TIMEOUT:
        _sendHungUpMessage(userId, {});

        _closeConnectionForOpponent(userId, (userId) {
          if (onUserNoAnswer != null) {
            onUserNoAnswer(this, userId);
          }
        });

        break;
      case PeerConnectionState.RTC_CONNECTION_CLOSED:
        break;

      case PeerConnectionState.RTC_CONNECTION_FAILED:
        _closeConnectionForOpponent(userId, (userId) {});
        break;
    }
  }

  void processHungUpFromUser(
    CubeRTCSessionDescription cubeRtcSdp,
    int opponentId,
  ) {
    log(
      "processHungUpFromUser, "
      "sessionId = ${cubeRtcSdp.sessionId}, "
      "senderId = $opponentId",
      TAG,
    );

    _closeConnectionForOpponent(opponentId, (opponentId) {
      if (onReceiveHungUpFromUser != null) {
        onReceiveHungUpFromUser(this, opponentId, _cubeSdp.userInfo);
      }
    });
  }

  void processRejectFromUser(
    CubeRTCSessionDescription cubeRtcSdp,
    int opponentId,
  ) {
    _closeConnectionForOpponent(opponentId, (opponentId) {
      if (onCallRejectedByUser != null) {
        onCallRejectedByUser(this, opponentId, _cubeSdp.userInfo);
      }
    });
  }

  void processIceCandidates(
    List<RTCIceCandidate> candidates,
    int opponentId,
  ) {
    log("processIceCandidates, _sdp = $_cubeSdp");
    PeerConnection peerConnection = _channels[opponentId];
    if (peerConnection == null) return;

    peerConnection.processIceCandidates(candidates);
  }

  Future<void> processAcceptCall(
    CubeRTCSessionDescription cubeRtcSdp,
    int opponentId,
    RTCSessionDescription sdp,
  ) async {
    PeerConnection peerConnection = _channels[opponentId];
    if (peerConnection == null) return;

    await peerConnection.processAnswer(sdp);
  }

  void _sendHungUpMessage(
    int opponentId,
    Map<String, String> userInfo,
  ) {
    MessageStanza hungUpMessage =
        createHungUpMessage(_cubeSdp, opponentId, userInfo);
    _client.sendSignalingMessage(hungUpMessage);
  }

  void _sendRejectMessage(
    int opponentId,
    Map<String, String> userInfo,
  ) {
    MessageStanza rejectMessage =
        createRejectMessage(_cubeSdp, opponentId, userInfo);
    _client.sendSignalingMessage(rejectMessage);
  }

  void processNewCall(
    int opponentId,
    RTCSessionDescription sdp,
  ) {
    PeerConnection peerConnection = _channels[opponentId];
    if (peerConnection == null) return;

    if (!peerConnection.hasRemoteSdp()) {
      peerConnection.setRemoteSdp(sdp);
    }

    if (_state == RTCSessionState.RTC_SESSION_CONNECTING ||
        _state == RTCSessionState.RTC_SESSION_CONNECTED) {
      peerConnection.startAnswer({});
    }
  }

  void _setState(RTCSessionState state) {
    if (this._state != state) {
      this._state = state;
    }
  }

  void _closeConnectionForOpponent(
    int opponentId,
    Function(int opponentId) callback,
  ) {
    PeerConnection peerConnection = _channels[opponentId];
    if (peerConnection == null) return;

    peerConnection.close();
    _channels.remove(opponentId);

    if (callback != null) {
      callback(opponentId);
    }

    log(
      "_closeConnectionForOpponent, "
      "_channels.length = ${_channels.length}",
      TAG,
    );

    if (_channels.length == 0) {
      _closeCurrentSession();
    } else {
      log(
        "_closeConnectionForOpponent, "
        "left channels = ${_channels.keys.join(", ")}",
        TAG,
      );
    }
  }

  void _closeCurrentSession() {
    log("_closeCurrentSession", TAG);
    if (_localStream != null) {
      _localStream.dispose();
      _localStream = null;
    }

    _notifySessionClosed();
  }

  void _notifySessionClosed() {
    log("_notifySessionClosed", TAG);
    _client.removeSession(this);

    if (onSessionClosed != null) {
      onSessionClosed(this);
    }
  }
}
