import 'dart:async';
import 'dart:math';

import 'package:flutter_webrtc/enums.dart';
import 'package:flutter_webrtc/media_stream.dart';
import 'package:flutter_webrtc/rtc_ice_candidate.dart';
import 'package:flutter_webrtc/rtc_peerconnection.dart';
import 'package:flutter_webrtc/rtc_peerconnection_factory.dart';
import 'package:flutter_webrtc/rtc_session_description.dart';

import 'package:connectycube_sdk/connectycube_chat.dart';

import 'utils/rtc_config.dart';
import 'utils/signaling_specifications.dart';

class PeerConnection {
  static const String TAG = "PeerConnection";
  int _userId;
  PeerConnectionStateCallback _peerConnectionStateCallback;

  RTCPeerConnection _peerConnection;
  List<RTCIceCandidate> _remoteCandidates = [];
  MediaStream _localMediaStream;
  RTCSessionDescription _remoteSdp;
  PeerConnectionState _signalingState;
  Timer _dillingTimer;
  int _startConnectionTime;

  PeerConnectionState get state => _signalingState;

  PeerConnection(this._userId, this._peerConnectionStateCallback) {
    _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_NEW);
  }

  Future<RTCPeerConnection> _createPeerConnection() async {
    _localMediaStream = this._peerConnectionStateCallback.getLocalMediaStream();

    RTCPeerConnection pc = await createPeerConnection(_iceServers, _config);
    pc.addStream(_localMediaStream);
    pc.onIceCandidate = (candidate) {
      this._peerConnectionStateCallback.sendIceCandidate(_userId, candidate);
    };

    pc.onIceConnectionState = (state) {
      log("onIceConnectionState changed to $state for opponent $_userId", TAG);

      if (RTCIceConnectionState.RTCIceConnectionStateChecking == state) {
        _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_CHECKING);
        _cancelDillingTimer();
      } else if (RTCIceConnectionState.RTCIceConnectionStateConnected ==
          state) {
        _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_CONNECTED);
      } else if (RTCIceConnectionState.RTCIceConnectionStateDisconnected ==
          state) {
        _onConnectionStateChanged(
            PeerConnectionState.RTC_CONNECTION_DISCONNECTED);
      } else if (RTCIceConnectionState.RTCIceConnectionStateFailed == state) {
        _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_FAILED);
      } else if (RTCIceConnectionState.RTCIceConnectionStateClosed == state) {
        _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_CLOSED);
      }
    };

    pc.onAddStream = (stream) {
      this._peerConnectionStateCallback.onRemoteStreamReceive(_userId, stream);
    };

    pc.onRemoveStream = (stream) {
      this._peerConnectionStateCallback.onRemoteStreamRemove(_userId, stream);
    };

    pc.onSignalingState = (state) {
      log("onSignalingState changed to $state for opponent $_userId", TAG);
    };

    return pc;
  }

  // TODO VT add possibility to use custom servers
  Map<String, dynamic> _iceServers = {
    'iceServers': [
      {'url': 'stun:turn.connectycube.com'},
      {
        'url': 'turn:turn.connectycube.com:5349?transport=udp',
        'username': 'connectycube',
        'credential': '4c29501ca9207b7fb9c4b4b6b04faeb1'
      },
      {
        'url': 'turn:turn.connectycube.com:5349?transport=tcp',
        'username': 'connectycube',
        'credential': '4c29501ca9207b7fb9c4b4b6b04faeb1'
      },
    ]
  };

  final Map<String, dynamic> _config = {
    'mandatory': {},
    'optional': [
      {'DtlsSrtpKeyAgreement': true},
    ],
  };

  Map<String, dynamic> get _constraints => {
        'mandatory': {
          'OfferToReceiveAudio': true,
          'OfferToReceiveVideo': _isVideoCall(),
        },
        'optional': [],
      };

  void startOffer(Map<String, String> userInfo) {
    _createPeerConnection().then((pc) {
      this._peerConnection = pc;

      _createOffer(pc, userInfo);
    });
  }

  _createOffer(
    RTCPeerConnection peerConnection,
    Map<String, String> userInfo,
  ) async {
    log("_createOffer for opponent $_userId", TAG);

    _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_PENDING);

    try {
      RTCSessionDescription sessionDescription =
          await peerConnection.createOffer(_constraints);
      peerConnection.setLocalDescription(sessionDescription);
      _sendOffer(sessionDescription, userInfo);
      _startDillingTimer(sessionDescription, userInfo);
    } catch (e) {
      log("_createOffer error: ${e.toString()}", TAG);
    }
  }

  startAnswer(Map<String, String> userInfo) async {
    if (PeerConnectionState.RTC_CONNECTION_PENDING != state &&
        PeerConnectionState.RTC_CONNECTION_NEW != state) return;

    log("startAnswer", TAG);

    _signalingState = PeerConnectionState.RTC_CONNECTION_CONNECTING;

    if (_remoteSdp == null) return;

    _peerConnection = await _createPeerConnection();
    await _peerConnection.setRemoteDescription(_remoteSdp);

    await _createAnswer(userInfo);

    if (this._remoteCandidates.length > 0) {
      _remoteCandidates.forEach((candidate) async {
        await _peerConnection.addCandidate(candidate);
      });
      _remoteCandidates.clear();
    }
  }

  _createAnswer(Map<String, String> userInfo) async {
    try {
      RTCSessionDescription sdp =
          await _peerConnection.createAnswer(_constraints);
      _peerConnection.setLocalDescription(sdp);
      _sendAnswer(sdp, userInfo);
    } catch (e) {
      log("_createAnswer error: ${e.toString()}", TAG);
    }
  }

  void _sendOffer(RTCSessionDescription sdp, Map<String, String> userInfo) {
    this._peerConnectionStateCallback.onSendOffer(_userId, sdp, userInfo);
  }

  void _sendAnswer(RTCSessionDescription sdp, Map<String, String> userInfo) {
    this._peerConnectionStateCallback.onSendAnswer(_userId, sdp, userInfo);
  }

  Future<void> processIceCandidates(List<RTCIceCandidate> iceCandidates) async {
    log("processIceCandidates, user id = $_userId", TAG);

    if (_peerConnection != null) {
      iceCandidates.forEach((candidate) async {
        log("processIceCandidates, ${await _peerConnection.getRemoteDescription()}",
            TAG);
        _peerConnection.addCandidate(candidate);
      });
    } else {
      _remoteCandidates.addAll(iceCandidates);
    }
  }

  void close() {
    _cancelDillingTimer();

    if (_peerConnection == null) return;

    _peerConnection.close();
    _peerConnection = null;
  }

  Future<void> processAnswer(RTCSessionDescription sdp) async {
    log("processAnswer, user id = $_userId", TAG);
    log("processAnswer, _remoteCandidates.lenght = ${_remoteCandidates.length}",
        TAG);

    if (_peerConnection != null) {
      log("processAnswer, _peerConnection != null", TAG);

//      RTCSessionDescription existRemoteSdp = await _peerConnection.getRemoteDescription();
//      if(existRemoteSdp == null) {
//        log("processAnswer, existRemoteSdp == null", TAG);
      await _peerConnection.setRemoteDescription(sdp);
//      } else {
//        log("processAnswer, existRemoteSdp != null", TAG);
//      }
    }
  }

  void setRemoteSdp(RTCSessionDescription sdp) {
    _remoteSdp = sdp;
  }

  bool _isVideoCall() {
    return CallType.VIDEO_CALL == _peerConnectionStateCallback?.callType;
  }

  void _startDillingTimer(
    RTCSessionDescription sessionDescription,
    Map<String, String> userInfo,
  ) {
    _startConnectionTime = DateTime.now().millisecondsSinceEpoch;
    _dillingTimer = Timer.periodic(
        Duration(
            seconds: max(RTCConfig.instance.defaultDillingTimeInterval,
                RTCConfig.instance.dillingTimeInterval)), (timer) {
      if (_isConnectionExpired()) {
        _onConnectionStateChanged(PeerConnectionState.RTC_CONNECTION_TIMEOUT);
        timer.cancel();
      } else {
        if (PeerConnectionState.RTC_CONNECTION_CONNECTING == state
            || PeerConnectionState.RTC_CONNECTION_PENDING == state) {
          _sendOffer(sessionDescription, userInfo);
        } else {
          timer.cancel();
        }
      }
    });
  }

  bool _isConnectionExpired() {
    int currentTime = DateTime.now().millisecondsSinceEpoch;
    int diff = currentTime - _startConnectionTime;
    bool isCallExpired = (diff / 1000) >= RTCConfig.instance.noAnswerTimeout;

    return isCallExpired;
  }

  void _onConnectionStateChanged(PeerConnectionState state) {
    _signalingState = state;
    _peerConnectionStateCallback.onPeerConnectionStateChanged(_userId, state);
  }

  void _cancelDillingTimer() {
    if (_dillingTimer != null) _dillingTimer.cancel();
  }

  bool hasRemoteSdp() {
    return _remoteSdp != null;
  }
}

abstract class PeerConnectionStateCallback {
  MediaStream getLocalMediaStream();

  int get callType;

  void onRemoteStreamReceive(
    int userId,
    MediaStream remoteMediaStream,
  );

  void onRemoteStreamRemove(
    int userId,
    MediaStream remoteMediaStream,
  );

  void sendIceCandidate(
    int userId,
    RTCIceCandidate iceCandidate,
  );

  void onSendOffer(
    int userId,
    RTCSessionDescription sdp,
    Map<String, String> userInfo,
  );

  void onSendAnswer(
    int userId,
    RTCSessionDescription sdp,
    Map<String, String> userInfo,
  );

  void onPeerConnectionStateChanged(
    int userId,
    PeerConnectionState state,
  );
}

enum PeerConnectionState {
  RTC_CONNECTION_NEW,
  RTC_CONNECTION_PENDING,
  RTC_CONNECTION_CONNECTING,
  RTC_CONNECTION_CHECKING,
  RTC_CONNECTION_CONNECTED,
  RTC_CONNECTION_DISCONNECTED,
  RTC_CONNECTION_TIMEOUT,
  RTC_CONNECTION_CLOSED,
  RTC_CONNECTION_FAILED
}
